// file : window.cpp

#include "../stdafx.h"
#include "../Resource/resource.h"

#include "window.h"

using namespace Core;

cWindow::cWindow(int _clientWidth, int _clientHeight)
{
	clientWidth = _clientWidth;
	clientHeight = _clientHeight;

	initializeWindow();
	initializeOpenGL();
}


cWindow::~cWindow()
{
	TCHAR className[32];
	LoadString(instance, ID_APP_CLASS_NAME, className, sizeof(className) / sizeof(TCHAR));
	::UnregisterClass(className, instance);
	
	wglMakeCurrent(nullptr, nullptr);
	wglDeleteContext(context);
	ReleaseDC(hwnd, device);
}


bool cWindow::initializeWindow()
{
	instance = (HINSTANCE)GetModuleHandle(NULL);
	
	// pull from string table resource
	TCHAR className[32];
	TCHAR title[32];
	LoadString(instance, ID_APP_TITLE, title, sizeof(title) / sizeof(TCHAR));
	LoadString(instance, ID_APP_CLASS_NAME, className, sizeof(className) / sizeof(TCHAR));

	WNDCLASSEX wcex;
	wcex.cbClsExtra = 0;
	wcex.cbSize = sizeof(WNDCLASSEX);
	wcex.cbWndExtra = sizeof(cWindow*);
	wcex.hbrBackground = (HBRUSH)GetStockObject(NULL_BRUSH);
	wcex.hCursor = LoadCursor(nullptr, IDC_ARROW);
	wcex.hIcon = LoadIcon(instance, MAKEINTRESOURCE(IDI_APP_ICON));
	wcex.hIconSm = LoadIcon(instance, MAKEINTRESOURCE(IDI_APP_ICON));
	wcex.hInstance = instance;
	wcex.lpfnWndProc = (WNDPROC)cWindow::messageProcedure;
	wcex.lpszClassName = className;
	wcex.lpszMenuName = nullptr; // no menu bar
	wcex.style = CS_OWNDC | CS_HREDRAW | CS_VREDRAW;

	if (!RegisterClassEx(&wcex))
	{
		Log("Error", "Failed to register window class");
		return false;
	}

	// window mode 
	// adjust creation dimensions for the addition of the window frame and title bar
	DWORD style = WS_SYSMENU | WS_CAPTION;
	RECT rectClient = { 0, 0, clientWidth, clientHeight };
	AdjustWindowRect(&rectClient, style, FALSE);
	UINT windowWidth = rectClient.right - rectClient.left;
	UINT windowHeight = rectClient.bottom - rectClient.top;

	// todo : leave this for now but consider not needing to store this and instead call system when needed.
	//        The stored dims would need updating when the window is moved
	rectWindow.left = (GetSystemMetrics(SM_CXSCREEN) / 2) - (windowWidth / 2);
	rectWindow.top  = (GetSystemMetrics(SM_CYSCREEN) / 2) - (windowHeight / 2);
	rectWindow.right = rectWindow.left + windowWidth;
	rectWindow.bottom = rectWindow.top + windowHeight;
	hwnd = CreateWindow(wcex.lpszClassName, title, style,
	   	                rectWindow.left, rectWindow.top, windowWidth, windowHeight,
		                nullptr, nullptr, instance, nullptr);

	if (!hwnd)
	{
		Log("Error", "Failed to create window");
		return false;
	}

	// save this class instance to the window extra handle
	::SetWindowLongPtr(hwnd, 0, reinterpret_cast<LONG_PTR>(this));
	return true;
}


bool cWindow::initializeOpenGL()
{
	device = GetDC(hwnd);

	PIXELFORMATDESCRIPTOR pfd;
	ZeroMemory(&pfd, sizeof(PIXELFORMATDESCRIPTOR));
	pfd.nSize = sizeof(PIXELFORMATDESCRIPTOR);
	pfd.nVersion = 1;
	pfd.dwFlags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
	pfd.iPixelType = PFD_TYPE_RGBA;
	pfd.cColorBits = 32;
	pfd.cDepthBits = 24;
	pfd.cStencilBits = 8;
	pfd.iLayerType = PFD_MAIN_PLANE;

	int format = ChoosePixelFormat(device, &pfd);
	if (format == 0)
	{
		Log("Error", "Failed to choose a pixel format");
		return false;
	}

	if (!SetPixelFormat(device, format, &pfd))
	{
		Log("Error", "Failed to set pixel format");
		return false;
	}

	context = wglCreateContext(device);
	if (!wglMakeCurrent(device, context))
	{
		Log("Error", "Failed to create render context");
		return false;
	}

	if (gl3wInit())
	{
		Log("Error", "Failed to initialize gl3w");
		return false;
	}

	//pLog(gl3wIsSupported(3, 3) ? Console::eMsgID::msg_normal : Console::eMsgID::msg_error, "OpenGL %s initialized", glGetString(GL_VERSION));
	//pLog(Console::eMsgID::msg_normal, "Renderer : %s %s\n", glGetString(GL_VENDOR), glGetString(GL_RENDERER));


	// set opengl state ----------------------------------------------
	glViewport(0, 0, clientWidth, clientHeight);

	glEnable(GL_DEPTH_TEST);
	glEnable(GL_TEXTURE_2D);

	//glCullFace(GL_BACK);
	//glEnable(GL_CULL_FACE);

	glClearColor(0.0f, 0.1f, 0.0f, 1.0f);

	//glPolygonMode(GL_FRONT_AND_BACK, GL_LINE); // wireframe mode
	glPolygonMode(GL_FRONT, GL_FILL);
	return true;
}


LRESULT cWindow::messageProcedure(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch (msg)
	{
	case WM_CLOSE:
		DestroyWindow(hwnd);
		return 0;

	case WM_DESTROY:
		PostQuitMessage(0);
		return 0;

	case WM_INPUT:
	{
		UINT nSize;
		if (GetRawInputData((HRAWINPUT)lParam, RID_INPUT, nullptr, &nSize, sizeof(RAWINPUTHEADER)) == -1)
			break;
		LPBYTE lpb = new BYTE[nSize];
		if (lpb == nullptr)
			return 0;

		if (GetRawInputData((HRAWINPUT)lParam, RID_INPUT, lpb, &nSize, sizeof(RAWINPUTHEADER)) != nSize)
			break; // silently fail		pCore->log("Warning", "GetRawInputData did not return expected size");

		RAWINPUT* raw = (RAWINPUT*)lpb;
		cWindow* wnd = reinterpret_cast<cWindow*>(GetWindowLongPtr(hwnd, 0));

		if (raw->header.dwType == RIM_TYPEKEYBOARD)
		{
			if (wnd != nullptr)
			{
				switch (raw->data.keyboard.Message)
				{ // store, remove or hold key presses
				case WM_KEYDOWN:
				{
					bool itemFound = false;
					for (auto item : wnd->keysActive)
					{
						if (item == raw->data.keyboard.VKey)
						{
							itemFound = true;
							break;
						}
					}
					if (!itemFound)
						wnd->keysActive.push_back(raw->data.keyboard.VKey);
					break;
				}

				case WM_KEYUP:
					wnd->keysActive.erase(std::remove(wnd->keysActive.begin(), wnd->keysActive.end(), raw->data.keyboard.VKey), wnd->keysActive.end());
					wnd->keysDeactivate.push_back(raw->data.keyboard.VKey);

				} // end switch
			}
		}

		else if (raw->header.dwType == RIM_TYPEMOUSE)
		{
			if (wnd != nullptr)
			{
				switch (raw->data.mouse.usButtonFlags)
				{
				case RI_MOUSE_LEFT_BUTTON_DOWN:
				{
					bool itemFound = false;
					for (auto item : wnd->keysActive)
					{
						if (item == VK_LBUTTON)
						{
							itemFound = true;
							break;
						}
					}
					if (!itemFound)
						wnd->keysActive.push_back(VK_LBUTTON);
					break;
				}
				case RI_MOUSE_RIGHT_BUTTON_DOWN:
				{
					bool itemFound = false;
					for (auto item : wnd->keysActive)
					{
						if (item == VK_RBUTTON)
						{
							itemFound = true;
							break;
						}
					}
					if (!itemFound)
						wnd->keysActive.push_back(VK_RBUTTON);
					break;
				}
				case RI_MOUSE_LEFT_BUTTON_UP:
					wnd->keysActive.erase(std::remove(wnd->keysActive.begin(), wnd->keysActive.end(), VK_LBUTTON), wnd->keysActive.end());
					wnd->keysDeactivate.push_back(VK_LBUTTON);
					break;
				case RI_MOUSE_RIGHT_BUTTON_UP:
					wnd->keysActive.erase(std::remove(wnd->keysActive.begin(), wnd->keysActive.end(), VK_RBUTTON), wnd->keysActive.end());
					wnd->keysDeactivate.push_back(VK_RBUTTON);
					break;
				} // end switch				
				
				wnd->mouseX = raw->data.mouse.lLastX;
				wnd->mouseY = raw->data.mouse.lLastY;
			}
		}

		delete[] lpb;
		break; // allow default handler to clean up
	}

	}
	return DefWindowProc(hwnd, msg, wParam, lParam);
}