// file : trueType.cpp

#include "../../stdafx.h"
#include "font.h"

#ifdef _DEBUG
#pragma comment(lib, "freetype-2.8.1/objs/vc2010/Win32/freetype281d.lib")
#else
#pragma comment(lib, "freetype-2.8.1/objs/vc2010/Win32/freetype281.lib")
#endif

using namespace Graphics::Text;

void cFont::initialize(std::string _fontName, unsigned int _height, glm::vec2 _screenDims)
{
	height = _height + 1; // pixel

	FT_Library library;
	FT_Face    face;
	// load library
	if (FT_Init_FreeType(&library))
	{
		Log("Error", "Failed to initialize FreeType library");
		return;
	}
	// load font
	std::string fontPath = "Assets/Fonts/" + _fontName;
	FT_Error error = FT_New_Face(library, fontPath.c_str(), 0, &face);
	if (error == FT_Err_Unknown_File_Format)
	{
		Log("Error", "Font file '%s' exists but format is unsupported", _fontName.c_str());
		return;
	}
	else if (error)
	{
		Log("Error", "Font file '%s' could not be opened", _fontName.c_str());
		return;
	}

	if (FT_Set_Pixel_Sizes(face, 0, height))
	{
		Log("Error", "Font file '%s' could not set pixel height of %d", _fontName.c_str(), height);
		return;
	}


	// build textures from glyphs
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1); // disable byte alignment restriction

	for (GLubyte ch = 0; ch < 128; ch++)
	{
		error = FT_Load_Char(face, ch, FT_LOAD_RENDER);
		if (error)
		{
			continue; // ignore errors 
		}

		GLuint texID;
		glGenTextures(1, &texID);
		glBindTexture(GL_TEXTURE_2D, texID);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RED, face->glyph->bitmap.width, face->glyph->bitmap.rows, 0, GL_RED, GL_UNSIGNED_BYTE, face->glyph->bitmap.buffer);

		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

		sCharacter character = { texID, glm::ivec2(face->glyph->bitmap.width, face->glyph->bitmap.rows), glm::ivec2(face->glyph->bitmap_left, face->glyph->bitmap_top), face->glyph->advance.x };
		characters.insert(std::pair<GLchar, sCharacter>(ch, character));
	}

	FT_Done_Face(face);
	FT_Done_FreeType(library);

	// single character dynamic quad setup
	glGenVertexArrays(1, &vao);
	glGenBuffers(1, &vbo);
	glBindVertexArray(vao);
	glBindBuffer(GL_ARRAY_BUFFER, vbo);
	glBufferData(GL_ARRAY_BUFFER, sizeof(GLfloat) * 6 * 4, NULL, GL_DYNAMIC_DRAW);

	glEnableVertexAttribArray(0);
	glVertexAttribPointer(0, 4, GL_FLOAT, GL_FALSE, 4 * sizeof(GLfloat), (GLvoid*)0);
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindVertexArray(0);

	projection = glm::ortho(0.0f, _screenDims.x, 0.0f, _screenDims.y);
}


void cFont::shutdown()
{
}


void cFont::enableState(GLuint _programID)
{
	glDisable(GL_DEPTH_TEST);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	programID = _programID;
	glUniformMatrix4fv(glGetUniformLocation(_programID, "projection"), 1, GL_FALSE, glm::value_ptr(projection));
	glActiveTexture(GL_TEXTURE0);
}


void cFont::disableState()
{
	glBindTexture(GL_TEXTURE_2D, 0);
	glEnable(GL_DEPTH_TEST);
	glDisable(GL_BLEND);
}


void cFont::draw(std::string _text, glm::vec3 _color)
{
	glUniform3f(glGetUniformLocation(programID, "fontColor"), _color.x, _color.y, _color.z);
	glBindVertexArray(vao);

	std::string::const_iterator iter;
	for (iter = _text.begin(); iter != _text.end(); ++iter)
	{
		if (*iter == '\n')
		{   // new line character
			cursorPos.x = origin.x;
			cursorPos.y -= height;
			continue;
		}

		if (*iter == '\t')
		{   // tab character
			cursorPos.x += (characters[' '].advance >> 6) * 4;
			continue;
		}

		sCharacter ch = characters[*iter];
		GLfloat x = cursorPos.x + ch.bearing.x;
		GLfloat y = cursorPos.y - (ch.size.y - ch.bearing.y);
		GLfloat w = (GLfloat)ch.size.x;
		GLfloat h = (GLfloat)ch.size.y;

		GLfloat vertices[6][4] = {
			{ x,     y + h, 0.0f, 0.0f },
			{ x,     y,     0.0f, 1.0f },
			{ x + w, y,     1.0f, 1.0f },
			{ x,     y + h, 0.0f, 0.0f },
			{ x + w, y,     1.0f, 1.0f },
			{ x + w, y + h, 1.0f, 0.0f }
		};

		glBindTexture(GL_TEXTURE_2D, ch.textureID);
		glBindBuffer(GL_ARRAY_BUFFER, vbo);
		glBufferSubData(GL_ARRAY_BUFFER, 0, sizeof(vertices), vertices);
		glBindBuffer(GL_ARRAY_BUFFER, 0);

		glDrawArrays(GL_TRIANGLES, 0, 6);
		cursorPos.x += (ch.advance >> 6);
	}
	glBindVertexArray(0);
}
